/**
  Shadow mapping demo application.
  
  Orion Sky Lawlor, olawlor@acm.org, 2010-04-15 (Public Domain)
*/
#include <iostream>
#include <cmath>
#include <GL/glew.h> /* 
OpenGL Extensions Wrangler:  for gl...ARB extentions.  Must call glewInit after glutCreateWindow! */
#include <GL/glut.h> /* OpenGL Utilities Toolkit, for GUI tools */
#include "ogl/glsl.h"
#include "ogl/glsl.cpp"
#include "ogl/minicam.h"
#include "osl/mat4.h"
#include "osl/mat4_inverse.h"

#include <vector>


class triangle {
public:
	int a[3]; // 1-based index into points and normals array
	triangle(int *a_) {
		for (int i=0;i<3;i++) a[i]=a_[i];
	}
};


class object {
public:
	std::vector<vec3> points; 
	std::vector<vec3> normals; 
	std::vector<vec3> texcoords; 
	std::vector<triangle> tris;
	
	object(const char *obj_file)
	{
		points.push_back(vec3(-999.0));
		normals.push_back(vec3(-999.0));
		texcoords.push_back(vec3(-999.0));
		FILE *f=fopen(obj_file,"r");
		char line[1000];
		while (NULL!=fgets(line,1000,f)) {
			switch (line[0]) {
			case 'v': {
				float x,y,z;
				sscanf(line,"v%f%f%f",&x,&y,&z);
				vec3 p=0.2*vec3(x,y,z); p.y-=0.4; p*=1.5;
				texcoords.push_back(p);
				points.push_back(p);
				normals.push_back(vec3(0.0,0.0,0.0));
				break;
			}
			case 'f': {
				int a[3];
				sscanf(line,"f%d%d%d",&a[0],&a[1],&a[2]);
				tris.push_back(triangle(a));
				break;
			}
			default: /* ignore that line */ ;
			}
		}
		fclose(f);
	}

	void perturb(float time) 
	{
		for (int p=0;p<points.size();p++)
		{
			points[p].x+=0.02*sin(10.0*points[p].x + time); // rippling fur
			normals[p]=vec3(0.0);
		}
		for (int f=0;f<tris.size();f++)
		{
			triangle &t=tris[f];
			vec3 N=normalize(cross(points[t.a[0]]-points[t.a[1]],
				points[t.a[0]]-points[t.a[2]]));
			for (int i=0;i<3;i++) normals[t.a[i]]+=N;
		}
		for (int n=1;n<(int)normals.size();n++) normals[n]=normalize(normals[n]);
	}

	void draw(float expanddist=0.0)
	{
		glBegin(GL_TRIANGLES); 
		for (int t=0;t<(int)tris.size();t++)
		{
			for (int i=0;i<3;i++) {
				glColor3ub(t&0xff,(t>>8)&0xff,(t>>16)&0xff);
				int vertex=tris[t].a[i];
				// The normal is used for lighting
				glNormal3fv(normals[vertex]);

				// The texcoords determine the hair roots
				vec3 p=texcoords[vertex];
				glTexCoord4fv(vec4(p.x,p.y,p.z,expanddist));

				// The geometry shows where stuff lands onscreen
				p=points[vertex];
				p+=expanddist*normals[vertex];
				glVertex3fv(p);
			}
		}
		glEnd();
	}
};

void rotate_cat(float time_uniform)
{
	glRotatef(time_uniform*10.0,0.3-sin(time_uniform*0.1),0.2,1.0);
	glTranslatef(1.0,0.0,0.0);
}
void display(void) 
{
	glEnable(GL_DEPTH_TEST);
	glDepthMask(GL_TRUE); // turn on Z writes
	glEnable(GL_BLEND);
	
	float time_uniform=glutGet(GLUT_ELAPSED_TIME)*0.001;
	
	/* Load up the geometry to render */
	static object obj_orig("cat_lopoly.obj");
	object obj=obj_orig;
	obj.perturb(time_uniform);
	
	
/*********** Draw from Light Source's View **********/
	/* Draw pixels with color == z */
	GLhandleARB zshader=MAKE_PROGRAM_OBJECT(
		varying vec3 position; // location of geometry
		void main(void) {
			vec4 world=gl_ModelViewMatrix*gl_Vertex;
			position=vec3(world);
			gl_Position=gl_ProjectionMatrix*world;
		}
	,
		varying vec3 position; // location of geometry
		void main(void) {
			gl_FragColor=vec4(vec3(position.z*0.5+0.5),1.0);
		}
	);
	glUseProgramObjectARB(zshader);
	
	int sw=4096,sh=4096; // pixels in shadow map	
	
	/* This is the shadowmap texture */
	static GLuint shadowmap=0;
	if (shadowmap==0) 
	{ /* only create texture once; it stays set up. */
		glGenTextures(1,&shadowmap);
		glBindTexture(GL_TEXTURE_2D,shadowmap);
		glTexImage2D(GL_TEXTURE_2D,0, // level 0 mipmap
			//GL_LUMINANCE8, //<- greyscale 8-bit image
			GL_RGBA8, //<- color 8-bit image
			//GL_RGB10_A2, //<- color 10-bit image
			//GL_RGBA32F_ARB, //<- floating-point pixels (fail!?)
			sw,sh,0,
			GL_RGBA,GL_FLOAT,0);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR); // no mipmaps
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
		glBindTexture(GL_TEXTURE_2D,0);
	}
	
	/* This framebuffer lets you render directly to the shadow map */
	static GLuint shadowfb=0;
	if (shadowfb==0) {
		glGenFramebuffersEXT(1, &shadowfb);
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, shadowfb);
		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, 
			GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, shadowmap, 0);
		
	}
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, shadowfb);
	glViewport(0,0,sw,sh);
	glClearColor(0,0,0,0); // shadow map version
	glClear(GL_COLOR_BUFFER_BIT);
	
	glMatrixMode(GL_PROJECTION); // no projection matrix-- -1 to +1
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	// Set up matrix to fit world coordinates into shadow map
	float s=0.5;
	glScalef(s,s,s);
	glTranslatef(0.3,0.0,0.0);
	// glRotatef(50*time_uniform,1,0,1);
	
	float shadowmat[4][4]; // transforms world into shadow coords
	glGetFloatv(GL_MODELVIEW_MATRIX,&shadowmat[0][0]);
	
	/* Draw using the GL_MAX blending mode: no depth buffer, but
	   we only keep the brightest color channels rendered. */
	glBlendEquation(GL_MAX);
	glBlendFunc(GL_ONE,GL_ONE);
	glEnable(GL_BLEND);
	glDisable(GL_DEPTH_TEST);
	
	rotate_cat(time_uniform);
	obj.draw(0); // draw to shadow map
	
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0); // back to screen display
	
	glUseProgramObjectARB(0);
	
	
/*********** Draw from Camera View **********/
	glViewport(0,0,glutGet(GLUT_WINDOW_WIDTH),glutGet(GLUT_WINDOW_HEIGHT));
	glClearColor(0.4,0.5,0.7,0.0); // background color
	glClearDepth(1.0); // normal camera version
	if (key_down['c']) // skip color buffer clear
		glClear(GL_DEPTH_BUFFER_BIT);
	else
		glClear(GL_COLOR_BUFFER_BIT+GL_DEPTH_BUFFER_BIT);
	
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LESS);
	glDisable(GL_BLEND);
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity(); // flush any ancient matrices
	gluPerspective(70.0, // fov
		glutGet(GLUT_WINDOW_WIDTH)/(float)glutGet(GLUT_WINDOW_HEIGHT),
		0.1,
		1000.0); // z clipping planes
	float cameraVelocity=3.0;
	oglCameraLookAt(cameraVelocity);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity(); // flush any ancient matrices
	
/* Upload 3D noise texture, to texture unit 2 */
	static GLuint noisetex=0;
	if (noisetex==0) { /* only need to build and upload one time; it stays set up. */
		glGenTextures(1,&noisetex);
		glBindTexture(GL_TEXTURE_3D,noisetex);
		int sz=32; /* RGBA 32 x 32 x 32 */
		float *noise=new float[4*sz*sz*sz];
		for (int z=0;z<sz;z++) 
		for (int y=0;y<sz;y++) 
		for (int x=0;x<sz;x++) 
		for (int chan=0;chan<4;chan++) {
			float r=(rand()&0xfff)*(1.0/0xfff);
			noise[chan+4*(x+sz*(y+sz*z))]=r;
		}
		glTexImage3D(GL_TEXTURE_3D,0,GL_RGBA8,
			sz,sz,sz,0,
			GL_RGBA,GL_FLOAT,noise);
		glTexParameteri(GL_TEXTURE_3D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
		glTexParameteri(GL_TEXTURE_3D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
		delete[] noise;
		glBindTexture(GL_TEXTURE_3D,0);
	}
	glActiveTexture(GL_TEXTURE2);
	glBindTexture(GL_TEXTURE_3D,noisetex);
	glActiveTexture(GL_TEXTURE1);
	glBindTexture(GL_TEXTURE_2D,shadowmap);
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D,0);

/* Full-on cat shader */
	GLhandleARB catshader=MAKE_PROGRAM_OBJECT(
		varying vec3 position; // location of fur-deformed geometry
		varying vec3 texcoords; // location of *original* geometry
		varying vec4 fg;
		varying vec4 bg; // vec4 colors (w/alpha)  
		varying vec3 normal; // surface normal
		void main(void) {
			vec4 world=gl_ModelViewMatrix*gl_Vertex;
			position=vec3(world);
			texcoords=vec3(gl_MultiTexCoord0);
			normal=vec3(gl_NormalMatrix*gl_Normal);
			gl_Position=gl_ProjectionMatrix*world;
			float fur=gl_MultiTexCoord0.w;
			float a=1.0-fur;
			fg=vec4(0.9,0.5,0.3,a); // light wood
			bg=vec4(0.3,0.2,0.1,a*0.2); // dark wood
			if (fur==0) {/* first layer */
				fg.a=bg.a=1.0;
			}
		}
	,
		varying vec3 position; // location of fur-deformed geometry
		varying vec3 texcoords; // location of *original* geometry
		varying vec4 fg;
		varying vec4 bg; // vec4 colors (w/alpha)
		varying vec3 normal; // surface normal
		uniform float time_uniform;
		
		uniform sampler3D noisetex;
		// Return value noise at this location
		vec4 mynoise(vec3 loc) {
			return texture3D(noisetex,loc);
		}
		
		// Return fur color
		vec4 furtexture(vec3 pos) {
			float blend=mynoise(pos*0.3).x;
			vec4 ret=bg+clamp(blend,0.0,1.0)*(fg-bg); // clamp color blend
			ret.a=bg.a+blend*(fg.a-bg.a); // don't clamp alpha blend--leaves spaces between hairs
			return ret;
		}
		
		uniform sampler2D shadowtex;
		uniform mat4 shadowmat;
		
		// Return the amount of light reaching this point
		vec3 light_at_point(vec3 loc) {
			vec3 shtex=vec3(shadowmat*vec4(loc,1.0));
			shtex=shtex*0.5+0.5; // -1 to +1 -> 0 to 1
			vec4 shval=texture2D(shadowtex,vec2(shtex));
			float fudge=1.0/256.0; // cures shadow acne!
			if (shtex.z<shval.z-fudge) return vec3(0.0); // in shadow
			else return vec3(1.0);
		}
		
		void main(void) {
			vec4 color=furtexture(texcoords*10.0);
		
			// Lighting calculation
			const vec3 L=normalize(vec3(0.0,0.0,1.0));
			vec3 N=normalize(normal);
			float d=clamp(dot(N,L),0.0,1.0);
			vec3 light=light_at_point(position);
			gl_FragColor=color*vec4(vec3(d)*light+0.3,1.0);
		}
	);
	glUseProgramObjectARB(catshader);
	glFastUniform1i(catshader,"noisetex",2); // texture unit 2
	glFastUniform1i(catshader,"shadowtex",1); // texture unit 1
	glFastUniformMatrix4fv(catshader,"shadowmat",1,0,&shadowmat[0][0]);
	glFastUniform1f(catshader,"time_uniform",time_uniform);
	
	glPushMatrix();
	rotate_cat(time_uniform);
	obj.draw(0.0); // draw to camera
	glPopMatrix();
	
	/* Also draw a rectangle on the ground */
	glNormal3f(0,0,1);
	glColor3f(1,1,1);
	glBegin(GL_TRIANGLE_FAN);
	float z=-1;
	glVertex3f(-5,-5,z);
	glVertex3f(+5,-5,z);
	glVertex3f(+5,+5,z);
	glVertex3f(-5,+5,z);
	glEnd();
	
	glUseProgramObjectARB(0);
	

/* Debug: draw boundary of shadow map */
	glColor3f(1,1,1);
	for (float z=-1;z<=+1;z+=1.0)
	{
		glBegin(GL_LINE_LOOP);
		glVertex3f(-1,-1,z);
		glVertex3f(+1,-1,z);
		glVertex3f(+1,+1,z);
		glVertex3f(-1,+1,z);
		glEnd();
	}
	
	
	glutPostRedisplay(); // continual animation
	glutSwapBuffers();
	
/* Estimate framerate */
	static double last_time=0.0;
	static int framecount=0, last_framecount=0;
	framecount++;
	double cur_time=0.001*glutGet(GLUT_ELAPSED_TIME);
	if (cur_time>0.5+last_time) 
	{ // every half a second, estimate fps
		double time=cur_time-last_time;
		double time_per_frame=time/(framecount-last_framecount);
		char str[100];
		sprintf(str,"Shadow Renderer: %.1f fps, %.1f ms/frame",
			1.0/time_per_frame,1.0e3*time_per_frame);
		glutSetWindowTitle(str);
		last_framecount=framecount;
		last_time=cur_time;
	}
}

int main(int argc,char *argv[]) 
{
	glutInit(&argc,argv);
	glutInitDisplayMode(GLUT_RGBA + GLUT_DEPTH + GLUT_DOUBLE);
	glutInitWindowSize(800,600);
	glutCreateWindow("Demo Application");
	
	glewInit();
	
	glutDisplayFunc(display);
	oglCameraInit();
	camera=vec3(0.0,-2.0,0.0);
	correct_head_tilt=true; // allow free flight
	
	glutMainLoop();
	return 0;
}
